/* auth-unix2_chkpwd.c -- authentification routine for vlock,
 *                        the VT locking program for linux
 *                        using external unix2_chkpwd binary
 *
 * This code is copyright (C) 2014 Stefan Seyfried, based on auth-shadow
 * which is copyright (C) 2007 Frank Benkstein, and is free
 * software which is freely distributable under the terms of the
 * GNU General Public License version 2, included as the file COPYING in this
 * distribution.  It is NOT public domain software, and any
 * redistribution not permitted by the GNU General Public License is
 * expressly forbidden without prior written permission from
 * the author.
 *
 */

/* for crypt() */
#define _XOPEN_SOURCE

#ifndef __FreeBSD__
/* for asprintf() */
#define _GNU_SOURCE
#endif

#include <stdbool.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <errno.h>
#include "auth.h"
#include "prompt.h"

bool auth(const char *user, struct timespec *timeout)
{
  char *pwd = NULL;
  char *msg;
  struct passwd *pw = NULL;
  int pfd[2], status;
  pid_t pid;
  int result = false;

  /* format the prompt */
  if (asprintf(&msg, "%s's Password: ", user) < 0)
    return false;

  pw = getpwnam(user);
  if (! pw)
    goto out;

  if ((pwd = prompt_echo_off(msg, timeout)) == NULL)
    goto out;

  if (pipe(pfd) < 0) {
    perror("vlock: pipe() failed");
    goto out;
  }

  if ((pid = fork()) < 0) {
    perror("vlock: fork() failed");
    close(pfd[0]);
    close(pfd[1]);
    goto out;
  }

  if (pid == 0) {
     close(pfd[1]);
     if (pfd[0] != 0)
       dup2(pfd[0], 0);

     /* Helper is invoked as helper service-name [user] */
     // printf("calling '/sbin/unix2_chkpwd vlock %s'\n", pw->pw_name);
     execlp("/sbin/unix2_chkpwd", "/sbin/unix2_chkpwd", "vlock", pw->pw_name, NULL);
     /* we should not come here! */
     perror("vlock: execlp(/sbin/unix2_chkpwd)");
     exit(1);
  }

  close(pfd[0]);
  /* Write out password to helper process */
  write(pfd[1], pwd, strlen(pwd));
  close(pfd[1]);

  while (waitpid(pid, &status, 0) < 0) {
    if (errno == EINTR)
      continue;
    perror("vlock: waitpid() failed");
    goto out;
  }

  if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
    fprintf(stderr, "vlock: Authentication error\n"); /* no sleep needed, unix2_chkpwd already does that */
  else
    result = true;

out:
  /* free the password, free(NULL) is ok */
  free(pwd);

  /* free the prompt */
  free(msg);

  return result;
}
